import struct
import spcxbutcher.spc
import logging

class SPCXParser:
    def __init__( self, filename ):
        self._file = open( filename, 'rb' )
        self._done = False
        self._count = 0

    def _parse( self ):
        while not self._done:
            unit = self._file.read( spcxbutcher.spc.UNIT_SIZE )
            spcUnitCount, = struct.unpack( '<L', unit )
            spc = self._parseSPC( spcUnitCount )
            if spc is None:
                continue
            yield spc

        self._verifySPCNumber( spcUnitCount )

    def _parseSPC( self, spcUnitCount ):
        spc = spcxbutcher.spc.fromFile( spcUnitCount, self._file )
        if spc is None:
            self._done = True
            return
        self._count += 1
        return spc

    def _verifySPCNumber( self, lastUnitRead ):
        expectedSPCNumber = lastUnitRead
        if expectedSPCNumber != self._count:
            raise Exception( "expected {} SPCs but found {}".format( expectedSPCNumber, self._count ) )

    def __iter__( self ):
        return self._parse()



class InvalidDescriptor( Exception ):
    pass

class Descriptor:
    def __init__( self, rawData ):
        self._validate( rawData )
        self._parseDescriptor( rawData )

    def _validate( self, rawData ):
        bit = {}
        for position in [ 24, 25, 27 ]:
            value = ( 1 << position ) & rawData
            bit[ position ] = value >> position

        high4bits = 0xf0000000 & rawData
        valid = high4bits == 0xc0000000\
                and bit[ 24 ] == 1 \
                and bit[ 25 ] == 0 \
                and bit[ 27 ] == 0

        if not valid:
            raise InvalidDescriptor( 'Invalid descriptor {:08x}'.format( rawData ) )

    def _parseDescriptor( self, descriptor ):
        highestByte = descriptor >> 24
        self._raw = ( highestByte & 0b00000100 ) >> 2
        self._timePerBin = descriptor & 0x00ffffff

    @property
    def raw( self ):
        return self._raw

    @property
    def timePerBin( self ):
        return self._timePerBin



class DeadtimeFilter:
    def __init__( self, deadtime ):
        self._deadtime = deadtime
        self._previous = None

    def __call__( self, current ):
        if self._previous is None:
            self._previous = current
            return True

        if current.timestamp - self._previous.timestamp > self._deadtime:
            self._previous = current
            return True

        return False



import logging

class HighTime:
    def __init__( self ):
        self._hightime = 0
        self._changed = False

    def inspect( self, rawData ):
        self._changed = False
        if self._change( rawData ):
            self._changed = True
            self._update( rawData )

    @property
    def value( self ):
        return self._hightime

    def changed( self ):
        return self._changed

    def _change( self, rawData ):
        mark = rawData & 0xc0000000
        return mark == 0x40000000

    def _update( self, rawData ):
        newHightime = self._extractHightime( rawData )
        if ( newHightime - self._hightime ) != ( 1 << 24 ):
            logging.warning( 'high timestamp bits changed more than expected! from {:08x} to {:08x}'.format( self._hightime, newHightime ) )
        self._hightime = newHightime

    def _extractHightime( self, rawData ):
        hightimeBits = 0x3fffffff & rawData
        return hightimeBits << 24



class PerChannelFilter:
    def __init__( self, filters ):
        self._filters = filters

    def __call__( self, record ):
        filter = self._filters[ record.channel ]
        return filter( record )



import logging

class InvalidEventRecord( Exception ):
    pass

class Event:
    def __init__( self, rawData, hightime ):
        self.type = 'event'
        timestamp = rawData & 0x00ffffff
        self.lvttl = self._lvttl( rawData )
        self.gap = rawData >> 29
        self.timestamp = hightime + timestamp

    def __repr__( self ):
        return str( ( self.lvttl, self.timestamp, self.gap ) )

    @classmethod
    def verifyEventRecordHeader( self, rawData ):
        highestTwoBits = ( rawData & 0xc0000000 ) >> 30
        if highestTwoBits not in [ 0b00, 0b01 ]:
            raise InvalidEventRecord( 'invalid rawData record: {:08x}'.format( rawData ) )

    @property
    def channel( self ):
        return self.lvttl

    def _lvttl( self, event ):
        channel = ( event >> 24 ) & 0b00011111
        if channel < 3 or channel > 14:
            raise Exception( "unexpected channel value: {}".format( channel ) )
        if channel <= 10:
            return channel - 2
        else:
            return channel - 4

    def __eq__( self, _tuple ):
        return ( self.lvttl, self.timestamp, self.gap ) == _tuple



import struct
import logging
from spcxbutcher import descriptor
from spcxbutcher import event
from spcxbutcher import hightime

UNIT_SIZE = 4
UNIT_FORMAT = '<L'
assert struct.calcsize( UNIT_FORMAT ) == UNIT_SIZE

class _NoMoreSPCs( Exception ): pass

class SPC:
    def __init__( self, unitCount, file ):
        self._hightime = hightime.HighTime()
        content = file.read( UNIT_SIZE * unitCount )
        self._iterator = struct.iter_unpack( UNIT_FORMAT, content )
        self._parse()

    def _parse( self ):
        self._parseDescriptor()
        self._skipGarbageEvent()
        self._parseEvents()

    def _parseDescriptor( self ):
        try:
            rawData, = next( self._iterator )
            self._descriptor = descriptor.Descriptor( rawData )
        except StopIteration:
            raise _NoMoreSPCs()

    def _skipGarbageEvent( self ):
        next( self._iterator )

    def _parseEvents( self ):
        for rawData, in self._iterator:
            event.Event.verifyEventRecordHeader( rawData )
            self._hightime.inspect( rawData )
            if self._hightime.changed():
                continue
            yield event.Event( rawData, self._hightime.value )

    @property
    def raw( self ):
        return self._descriptor.raw

    @property
    def timePerBin( self ):
        return self._descriptor.timePerBin

    def __iter__( self ):
        return self._parseEvents()

def fromFile( unitCount, file ):
    try:
        return SPC( unitCount, file )
    except _NoMoreSPCs:
        return None



# spcxbutcher's setup.py
from distutils.core import setup
setup(
    name = "spcxbutcher",
    packages = ["spcxbutcher"],
    version = "0.2",
    description = "SPCX photon data parser",
    author = "Moran Netser & Yoav Kleinberger",
    author_email = "haarcuba@gmail.com",
    url = "https://github.com/morannetser/SpcxButcher",
    keywords = ["SPCX", "SPC", "photon", "photons"],
    scripts = ['bin/spcx_to_json', 'bin/spcx_to_matlab', 'bin/spcx_view'],
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Physics"
        ]
)
